/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you add "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * 24h timer with web interface.
 * A button connected to PB0 may be used to manually switch the relay.
 * 
 * See http://tuxgraphics.org/electronics/
 *
 * Chip type: Atmega328  or atmega644 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"
#include "websrv_help_functions.h"

// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)
//
#define RELAYON PORTD|=(1<<PORTD7)
#define RELAYOFF PORTD&=~(1<<PORTD7)
#define RELAYISON PORTD&(1<<PORTD7)

// please modify the following lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices. The IP address may be changed at run-time but the
// MAC address is hardcoded:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
//
// The IP of this device (can also be changed at run-time, see README file):
static uint8_t myip[4] = {10,0,0,29};
//static uint8_t myip[4] = {192,168,0,81};
// listen port for tcp/www:
static uint16_t mywwwport=80;
//static uint16_t mywwwport=81;
// the password string (only characters: a-z,0-9,_,.,-,# ), max len 8 char:
static char password[9]="secret";
// You can calibrate the clock here a bit.
// The number can be positive or negative and says
// how many seconds to add or subtract every 240h hours.
// I recommend to let the board run for a month before you
// touch this number.
#define CLOCK_CAL 1
// To make it easier to calibrate the system you can have the system 
// show seconds on the main page. Note that seconds jump once in a while.
//#define CLOCK_SHOW_SEC 
// -------------- do not change anything below this line ----------
// The buffer is the packet size we can handle and its upper limit is
// given by the amount of RAM that the atmega chip has. 
#define BUFFER_SIZE 980
static uint8_t buf[BUFFER_SIZE+1];

// global string buffer
#define STR_BUFFER_SIZE 40
static char gStrbuf[STR_BUFFER_SIZE+1];
//
static uint16_t gPlen;
static uint16_t tablebitmap[6]={0,0,0,0,0,0}; // each of the 96 bits represents one field in the time table
static volatile uint8_t time[3]={0,0,0}; // hour=0, min=1, sec=2
// 
uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,strlen(password))==0){
                return(1);
        }
        return(0);
}

uint16_t http200ok(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}

uint16_t http200okjs(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: application/x-javascript\r\n\r\n")));
}

uint16_t fill_tcp_data_uint8_pad(uint8_t *buf,uint16_t plen,uint8_t i)
{
        itoa(i,gStrbuf,10); // convert integer to string
        if (i<10) plen=fill_tcp_data(buf,plen,"0");
        return(fill_tcp_data(buf,plen,gStrbuf));
}

uint16_t fill_tcp_data_int(uint8_t *buf,uint16_t plen,uint16_t ui)
{
        uint8_t i=6;
        if (ui==0){
                return(fill_tcp_data(buf,plen,"0"));
        }
        // convert a 16bit unsigned integer into a printable string, avr-lib has
        // not function for that therefore we do it step by step:
        gStrbuf[i]='\0';
        while(ui&&i){
                i--;
                gStrbuf[i]=(char)((ui%10 & 0xf))+0x30;
                ui=ui/10;
        }
        return(fill_tcp_data(buf,plen,&gStrbuf[i]));
}

/***************
viewport tag properties (see as well
http://developer.apple.com/library/safari/#documentation/appleapplications/refer
+ence/SafariHTMLRef/Articles/InputTypes.html#//apple_ref/doc/uid/TP40008055-SW3
)

width - The width of the viewport in pixels. The default is 980. The
range is from 200 to 10,000.
You can as well set it to the special value "device-width"

height - The height of the viewport in pixels. The default is calculated
based on the value of the width property and the aspect ratio of the
device. The range is from 223 to 10,000 pixels.
You can as well set it to the special value "device-height"

initial-scale - The initial scale of the viewport as a multiplier. The
default is calculated to fit the webpage in the visible area. The range
is determined by the minimum-scale and maximum-scale properties.

You can set only the initial scale of the viewport?the scale of the
viewport the first time the webpage is displayed. Thereafter, the user
can zoom in and out unless you set user-scalable to no. Zooming by the
user is also limited by the minimum-scale and maximum-scale properties.

minimum-scale - Specifies the minimum scale value of the viewport. The
default is 0.25. The range is from >0 to 10.0.

maximum-scale - Specifies the maximum scale value of the viewport. The
default is 1.6. The range is from >0 to 10.0.

user-scalable - Determines whether or not the user can zoom in and
out?whether or not the user can change the scale of the viewport. Set to
yes to allow scaling and no to disallow scaling. The default is yes.

Setting user-scalable to no also prevents a webpage from scrolling when
entering text in an input field.

****************/
// there is unfortunately one problem: Differnet mobile browsers
// interprete the same syntax in different ways. Especially Opera
// is quite off from safari and the android browser. The only
// thing that works well is width=device-width
uint16_t print_line_mobile_webpage(uint16_t plen)
{
        //plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width, initial-scale=1.2\">\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width\">\n"));
        //plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=240, initial-scale=1\">\n"));
        return(plen);
}

uint16_t print_webpage_setclock(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Set clock</h2>\n<pre>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<form  action=cc method=get>\n\
time :<input type=text name=tv id=tf size=10>\n\
passw:<input type=password name=pw size=10><input type=submit value=\"set time\">\n\
</form>\n\
<script>\n\
d=new Date();\n\
e=document.getElementById(\"tf\");\n\
m=d.getMinutes();\n\
s= (m<10) ? \"0\" : \"\";\n\
e.value=d.getHours()+\":\"+s+d.getMinutes();\n\
</script>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre><hr>\n"));
        return(plen);
}

uint16_t print_webpage_relay(void)
{
        uint16_t plen;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a> <a href=/sw>[refresh]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Switch</h2>\n<pre>state: "));
        if (RELAYISON){
                plen=fill_tcp_data_p(buf,plen,PSTR("<font color=#00FF00>ON</font>"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("OFF"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\n<form action=/cr method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=sw value="));
        if (RELAYISON){
                plen=fill_tcp_data_p(buf,plen,PSTR("0>\npassw: <input type=password size=10 name=pw>"));
                plen=fill_tcp_data_p(buf,plen,PSTR("<input type=submit value=\"switch off\"></form>\n"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("1>\npassw: <input type=password size=10 name=pw>"));
                plen=fill_tcp_data_p(buf,plen,PSTR("<input type=submit value=\"switch on\"></form>\n"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("</pre><hr>\n"));
        return(plen);
}

// check the tablebitmap and return:
// 0 there is no change compared to previous minute and current state is off
// 1 the on/off state changed from on to off
// 2 the on/off state changed from off to on
// 3 there is no change compared to previous minute and current state is on
uint8_t chk_tablebitmap(uint8_t h, uint8_t m)
{
        uint8_t tablepage,bitpos;
        uint8_t curr=0;
        uint8_t prev=0;
        tablepage=h/4;// which table page is the current
        if (tablepage>5) tablepage=5; // bounds check in case of garbage time
        bitpos=((h%4)*60+m)/15;
        if (bitpos>15) bitpos=15; // bounds check in case of garbage time
        if (tablebitmap[tablepage] & (1<<bitpos)){
                curr=1;
        }
        // now check what the previous value was
        if (m>0){
                m--;
        }else{
                m=59;
                if (h>0){
                        h--;
                }else{
                        h=23;
                }
        }
        tablepage=h/4;// which table page is the current
        if (tablepage>5) tablepage=5; // bounds check in case of garbage time
        bitpos=((h%4)*60+m)/15;
        if (bitpos>15) bitpos=15; // bounds check in case of garbage time
        if (tablebitmap[tablepage] & (1<<bitpos)){
                prev=1;
        }
        if (prev==curr && curr==0) return(0);
        if (prev==curr && curr==1) return(3);
        if (curr==1 && prev==0) return(2);
        return(1);
}

//void relay_change_check(void){
//        uint8_t i;
//        i=chk_tablebitmap(time[0],time[1]);
//        if (i==2){
//                RELAYON;
//        }
//        if (i==1){
//                RELAYOFF;
//        }
//}

void relay_on_off_check(void){
        uint8_t i;
        i=chk_tablebitmap(time[0],time[1]);
        if (i==2||i==3){
                RELAYON;
        }else{
                RELAYOFF;
        }
}

// p1.js
uint16_t print_p1_js(void)
{
        // I am sorry that the javascript is a bit unreadable 
        // but we absolutely need to keep the amount of code
        // in bytes small.
        //
        // In the function fill we encode all the 96 fiels into 6 numbers 
        // with 16 bit. Every bit represents a table box. If the box was 
        // green then the bit is 1 otherwise zero.
        uint16_t plen;
        plen=http200okjs();
        plen=fill_tcp_data_p(buf,plen,PSTR("\
var col=\"\";\n\
var cc=0;\n\
var e,i,j;\n\
function dw(s){document.writeln(s)}\n\
function wr(t,i){dw(\"<tr><td>\"+t+\"</td><td id=t\"+i+\" onmousedown=cbg(this) onmouseover=mc(this)>&nbsp;</td></tr>\")}\n\
function dwh(s){dw(\"<table border=1><tr><td>Time</td><td>green=on<br>white=off</td></tr>\")}\n\
function cbg(e) {\n\
cc=1;\n\
col=e.style.backgroundColor ? \"\" : \"#7AFA7A\";\n\
e.style.backgroundColor=col;\n\
}\n\
function sc(){cc=0}\n\
function fill(ff)\n\
{\n\
var b=new Array();\n\
for(i=0;i<6;i++){\n\
b[i]=0;\n\
for(j=0;j<16;j++){\n\
e=document.getElementById(\"t\"+(i*16+j));\n\
if (e.style.backgroundColor) b[i]|=1<<j;\n\
}\n\
}\n\
ff.elements[0].value=b[0] +\"-\"+b[1]+\"-\"+b[2]+\"-\"+b[3]+\"-\"+b[4]+\"-\"+b[5]+\"-\";\n\
}\n\
function it(a)\n\
{\n\
for(i=0;i<6;i++){\n\
for(j=0;j<16;j++){\n\
e=document.getElementById(\"t\"+(i*16+j));\n\
if (a[i]&(1<<j)) e.style.backgroundColor=\"#7AFA7A\";\n\
}\n\
}\n\
}"));
        return(plen);
}


// this the top level main web page
uint16_t print_webpage_main(void)
{
        uint16_t plen;
        int8_t i;
        plen=http200ok();
        plen=print_line_mobile_webpage(plen);
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/sc>[set]</a> <a href=/sw>[switch]</a> <a href=/>[refresh]</a>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h1>24h Timer</h1>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("Current time:"));
        plen=fill_tcp_data_uint8_pad(buf,plen,time[0]);
        plen=fill_tcp_data(buf,plen,":");
        plen=fill_tcp_data_uint8_pad(buf,plen,time[1]);
#ifdef CLOCK_SHOW_SEC
        plen=fill_tcp_data(buf,plen,":");
        plen=fill_tcp_data_uint8_pad(buf,plen,time[2]);
#endif
        plen=fill_tcp_data_p(buf,plen,PSTR(" Output: "));
        if (RELAYISON){
                plen=fill_tcp_data_p(buf,plen,PSTR("on"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("off"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\
<br>Click table fields to change.\n\
<script src=p1.js></script>\n\
<script>\n\
function mc(e){if (cc) e.style.backgroundColor=col}\n\
dw(\"<table onmouseup=sc()>\");\n\
dw(\"<tr><td>\");\n\
dwh();\n\
var i,t;\n\
for(i=0;i<48;i++){\n\
t=\"\";\n\
if (i%4==0) t=i/4+\":00\";\n\
wr(t,i);\n\
}\n\
dw(\"</table></td><td>\");\n\
dwh();\n\
for(i=48;i<96;i++){\n\
t=\"\";\n\
if (i%4==0) t=i/4+\":00\";\n\
wr(t,i);\n\
}\n\
dw(\"</table></td></tr></table>\");\n\
dw(\"1 row=15min\");\n\
"));
        // initialize the table with the current settings
        plen=fill_tcp_data_p(buf,plen,PSTR("it(new Array("));
        plen=fill_tcp_data_int(buf,plen,tablebitmap[0]);
        i=1;
        while(i<6){
                plen=fill_tcp_data(buf,plen,",");
                plen=fill_tcp_data_int(buf,plen,tablebitmap[i]);
                i++;
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("));\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\
</script>\n\
<br><br>\n\
<form onsubmit=fill(this) action=st method=get>\n\
<input type=hidden name=tt>\n\
passw: <input type=password size=10 name=pw>\n\
<input type=submit value=save>\n\
</form>\n\
<hr>(c)tuxgraphics\n\
"));
        // debug dbg
        //plen=fill_tcp_data_int(buf,plen,chk_tablebitmap(time[0],time[1]));
        return(plen);
}

// takes a string of the form command/Number and analyse it (e.g "cc?tv=8%3A15&pw=secret HTTP/1.1")
// The first char of the url ('/') is already removed.
int8_t analyse_get_url(char *str)
{
        int8_t i;
        int8_t j;
        uint8_t n;
        if (str[0]==' '){ // end of url, main page
                return(1); 
        }
        // --------
        if (strncmp("favicon.ico",str,11)==0){
                gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("http://tuxgraphics.org/ico/clk.ico"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("\r\n\r\nContent-Type: text/html\r\n\r\n"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>301 Moved Permanently</h1>\n"));
                return(10);
        }
        // --------
        if (strncmp("sc",str,2)==0){
                gPlen=print_webpage_setclock();
                return(10);
        }
        if (strncmp("sw",str,2)==0){
                gPlen=print_webpage_relay();
                return(10);
        }
        if (strncmp("p1.js",str,5)==0){
                gPlen=print_p1_js();
                return(10);
        }
        // --------
        if (strncmp("st",str,2)==0){ // save the tablebitmap
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"tt")){
                                        urldecode(gStrbuf);
                                        gStrbuf[37]='\0'; // max length
                                        // gStrbuf is a string like this: 11-58-0-0-0-0-
                                        // no space allowed
                                        if (!isdigit(gStrbuf[0])){
                                                return(0);
                                        }
                                        n=0;i=0;j=0;
                                        while(gStrbuf[i] && n<6){
                                                if(gStrbuf[i]=='-'){
                                                        gStrbuf[i]='\0';
                                                        tablebitmap[n]=atol(&gStrbuf[j]);
                                                        n++;
                                                        j=i+1;
                                                }
                                                i++;
                                        }
                                        relay_on_off_check();
                                        return(1); // main page
                                }
                        }
                }
                return(-1);
        }
        if (strncmp("cc",str,2)==0){ // change clock
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"tv")){
                                        urldecode(gStrbuf);
                                        gStrbuf[5]='\0';
                                        // gStrbuf is a string like this: 11:58
                                        // no space allowed
                                        if (!isdigit(gStrbuf[0])){
                                                return(0);
                                        }
                                        // find position of ':' in string:
                                        i=0;
                                        if (gStrbuf[1]==':'){
                                                i=1;
                                        }
                                        if (gStrbuf[2]==':'){
                                                i=2;
                                        }
                                        if (i){
                                                gStrbuf[i]='\0';
                                                time[0]=atoi(gStrbuf);
                                                time[1]=atoi(&gStrbuf[i+1]);
                                                time[2]=0; 
                                                relay_on_off_check();
                                                return(1); // main page
                                        }
                                        return(0);
                                }
                        }
                }
                return(-1);
        }
        i=0; // switch on or off 
        if (strncmp("cr",str,2)==0){ // change relay state FORM
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"sw")){
                        if (gStrbuf[0]=='1'){
                                i=1;
                        }
                        if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                                urldecode(gStrbuf);
                                if (verify_password(gStrbuf)){
                                        if (i){
                                                RELAYON;
                                        }else{
                                                RELAYOFF;
                                        }
                                        return(1); // main page
                                }
                        }
                        return(-1);
                }
                return(0);
        }
        return(0);
}

// called when TCNT2==OCR2A
// that is in 50Hz intervals (49.8246Hz intervals, clock is a bit behind)
// This is used as a "clock" to store the data
ISR(TIMER2_COMPA_vect){
        static uint8_t cnt2step=0;
        cnt2step++;
        if (cnt2step>49){ 
                time[2]++; // step sec
                cnt2step=0;
                // toggle led:
                if (LEDISOFF){
                        LEDON;
                }else{
                        LEDOFF;
                }
        }
}

/* setup timer T2 as an interrupt generating time base.
* You must call once sei() in the main program */
void init_cnt2(void)
{
        PRR&=~(1<<PRTIM2); // write power reduction register to zero
        TIMSK2=(1<<OCIE2A); // compare match on OCR2A
        TCNT2=0;  // init counter
        OCR2A=244; // value to compare against (we divide by 245)
        TCCR2A=(1<<WGM21); // do not change any output pin, clear at compare match
        // divide clock by 1024: 12.5MHz/1024=12207.0313 Hz
        TCCR2B=(1<<CS22)|(1<<CS21)|(1<<CS20); // clock divider, start counter
        // 12207.0313 / 245= 49.8246Hz
}


int main(void){
        uint16_t dat_p;
        uint16_t plen;
        int8_t i=0;
        uint8_t clkchange=0;
        uint8_t newrelaystate=0;
        uint8_t days=0;
        uint16_t debounce=0;
        uint8_t gratutious_arp_cnt=0;
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        // enable PD6, as input, change state of relay manually
        DDRD&= ~(1<<DDD6);
        PORTD|= (1<<PIND6); // internal pullup resistor on, jumper goes to ground
        //initialize enc28j60
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us
        // Magjack leds configuration, see enc28j60 datasheet, page 11 
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        //
        // time keeping
        init_cnt2();
        sei();
        // LED
        DDRB|= (1<<DDB1); // enable PB1, LED as output
        LEDOFF;

        // the transistor on PD7 (relay)
        DDRD|= (1<<DDD7);
        RELAYOFF;

        //init the web server ethernet/ip layer:
        init_udp_or_www_server(mymac,myip);
        www_server_port(mywwwport);

        // initialize:
        while(1){
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);
                dat_p=packetloop_arp_icmp_tcp(buf,plen);

                //dat_p will ne unequal to zero if there is a valid packet (without crc error) 
                if(dat_p==0){
                        // no packets, do time keeping and relay control
                        if (newrelaystate>0){
                                if (newrelaystate==1){
                                        RELAYON;
                                }else{
                                        RELAYOFF;
                                }
                                newrelaystate=0;
                                // do nothing else but relay switchin in this cycle:
                                continue;
                        }
                        //
                        clkchange=0;
                        // if the crystal would be exact then we
                        // would be 303.0609 seconds behind every day.
                        // 24*3600 - 24*3600 * 0.9965=303.0609
                        // We compensate 12 sec every hour:
                        // 303.0609 - 12*24=303.0609 - 288=15.069
                        // That is every day we are 15.069 sec behind
                        // We compensate that at 23:30:00 with 15 sec
                        // That means we are 0.069 sec behind per day.
                        if (time[2]>59){ // sec
                                time[2]=0;
                                time[1]++; // count up min
                                clkchange=1; // marker: minutes have changed
                                // count gratutious_arp_cnt down every min:
                                if (gratutious_arp_cnt) gratutious_arp_cnt--;
                        }
                        if (time[1]>59){ // min
                                time[1]=0;
                                time[0]++; // count up hour
                                // We calibrate the clock every hour a bit:
                                // our clock is a bit behind.
                                time[2]+=12; // adjust/compensate
                        }
                        // once a day at 23:30
                        if (time[0]==23 && time[1] == 30 && time[2]<4){
                                time[2]+=15; // adjust/compensate
                        }
                        if (time[0]>23){ // hours
                                time[0]=0;
                                days++;
                        }
                        // every 10 days:
                        if (days==10){
                                time[2]+=CLOCK_CAL; // adjust/compensate 
                                days=0; // reset to zero otherwise we would do +CLOCK_CAL for a whole day
                        }
                        // Make this work with broken wlan/wifi equipment.
                        // Some of those wifi bridges or wifi client APs
                        // need once in a while a packet otherwise they
                        // forget the MAC address of the connected device and
                        // there is a flaw in those devices such that they
                        // do not broadcast in case of unknown MAC.
                        if (time[2]==15 && gratutious_arp_cnt==0){ // we use time[2], to have an inital arp shortly after power-on
                                gratutious_arp_cnt=5; // every 5min
                                gratutious_arp(buf);
                        }
                        //
                        if (clkchange){ // the values of the minutes did change
                                clkchange=chk_tablebitmap(time[0],time[1]);
                                // we do the actual relay switching 
                                // in the next idle cycle:
                                if (clkchange==2){
                                        newrelaystate=1; // on
                                }
                                if (clkchange==1){
                                        newrelaystate=2; // off
                                }
                        }
                        // a button connected between PD6 and GND to
                        // manually change the state of the relay:
                        if (debounce==0 && bit_is_clear(PIND,PIND6)){
                                if (RELAYISON){
                                        RELAYOFF;
                                }else{
                                        RELAYON;
                                }
                                debounce=0x4fff;
                        }
                        if (debounce) debounce--; // count up and wrap to zero
                        continue;
                }
                // Cut the size for security reasons. If we are almost at the 
                // end of the buffer then there is a zero but normally there is
                // a lot of room and we can cut down the processing time as 
                // correct URLs should be short in our case. If dat_p is already
                // close to the end then the buffer is terminated already.
                if ((dat_p+65) < BUFFER_SIZE){
                        buf[dat_p+65]='\0';
                }
                // analyse the url and do possible port changes:
                // move one char ahead:
                i=analyse_get_url((char *)&(buf[dat_p+5]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (i==-1){
                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1><a href=/>continue</a>\n"));
                        goto SENDTCP;
                }
                if (i==0){
                        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                        goto SENDTCP;
                }
                if (i==10){
                        plen=gPlen; // if return code 10 then data len is in gPlen
                        goto SENDTCP;
                }
                // just display the status:
                // normally i==1
                plen=print_webpage_main();
                //
SENDTCP:
                www_server_reply(buf,plen); // send web page data
                // tcp port www end
                //
        }
        return (0);
}
